"""This verifies zeekclient's ability to ingest cluster configurations, validate
their content (excluding deeper validations happening in the cluster
controller), and serialize them correctly to INI/JSON.
"""

import configparser
import io
import json
import unittest
from unittest.mock import MagicMock, patch

import zeekclient


class TestRendering(unittest.TestCase):
    INI_INPUT = """# A sample ini using all available keys.
[instances]
agent

[manager]
instance = agent
port = 5000
role = manager
metrics_port = 6000

[logger-01]
instance = agent
port = 5001
role = logger
scripts = foo/bar/baz

[worker-01]
instance = agent
role = worker
interface = lo
env = FOO=BAR BLUM=frub
cpu_affinity = 4

[worker-02]
instance = agent
role = worker
interface = enp3s0
cpu_affinity = 8
metrics_port = 6001
"""
    INI_EXPECTED = """[instances]
agent

[logger-01]
instance = agent
role = LOGGER
port = 5001
scripts = foo/bar/baz

[manager]
instance = agent
role = MANAGER
port = 5000
metrics_port = 6000

[worker-01]
instance = agent
role = WORKER
interface = lo
cpu_affinity = 4
env = BLUM=frub FOO=BAR

[worker-02]
instance = agent
role = WORKER
interface = enp3s0
cpu_affinity = 8
metrics_port = 6001
"""
    JSON_EXPECTED = """{
    "id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
    "instances": [
        {
            "name": "agent"
        }
    ],
    "nodes": [
        {
            "cpu_affinity": null,
            "env": {},
            "instance": "agent",
            "interface": null,
            "metrics_port": null,
            "name": "logger-01",
            "options": null,
            "port": 5001,
            "role": "LOGGER",
            "scripts": [
                "foo/bar/baz"
            ]
        },
        {
            "cpu_affinity": null,
            "env": {},
            "instance": "agent",
            "interface": null,
            "metrics_port": 6000,
            "name": "manager",
            "options": null,
            "port": 5000,
            "role": "MANAGER",
            "scripts": null
        },
        {
            "cpu_affinity": 4,
            "env": {
                "BLUM": "frub",
                "FOO": "BAR"
            },
            "instance": "agent",
            "interface": "lo",
            "metrics_port": null,
            "name": "worker-01",
            "options": null,
            "port": null,
            "role": "WORKER",
            "scripts": null
        },
        {
            "cpu_affinity": 8,
            "env": {},
            "instance": "agent",
            "interface": "enp3s0",
            "metrics_port": 6001,
            "name": "worker-02",
            "options": null,
            "port": null,
            "role": "WORKER",
            "scripts": null
        }
    ]
}"""

    def assertEqualStripped(self, str1, str2):  # noqa: N802
        self.assertEqual(str1.strip(), str2.strip())

    def parser_from_string(self, content):
        cfp = configparser.ConfigParser(allow_no_value=True)
        cfp.read_string(content)
        return cfp

    def setUp(self):
        # A buffer receiving any created log messages, for validation. We could
        # also assertLogs(), but with the latter it's more work to get exactly
        # the output the user would see.
        self.logbuf = io.StringIO()
        zeekclient.logs.configure(verbosity=3, stream=self.logbuf)

    def test_full_config_ini(self):
        # This test parses a feature-complete configuration from an INI file,
        # and verifies that writing it back out to an INI yields expected
        # content.

        # Parse the input into a config parser, and create a Configuration
        # object from it.
        cfp = self.parser_from_string(self.INI_INPUT)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertTrue(config is not None)

        # Turning that back into a config parser should have expected content:
        cfp = config.to_config_parser()
        with io.StringIO() as buf:
            cfp.write(buf)
            self.assertEqualStripped(buf.getvalue(), self.INI_EXPECTED)

        # Another roundtrip: the content should not change.
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertTrue(config is not None)

        cfp = config.to_config_parser()
        with io.StringIO() as buf:
            cfp.write(buf)
            self.assertEqualStripped(buf.getvalue(), self.INI_EXPECTED)

    def test_full_config_json(self):
        # This test parses a feature-complete configuration from an INI file,
        # and verifies that writing it to JSON yields expected content.

        # Parse the input into a config parser, and create a Configuration
        # object from it.
        cfp = self.parser_from_string(self.INI_INPUT)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertTrue(config is not None)

        jdata = config.to_json_data()

        def canon(c):
            """Canonicalize the ID"""
            return "-" if c == "-" else "x"

        jdata["id"] = "".join([canon(c) for c in jdata["id"]])

        self.assertEqual(
            json.dumps(jdata, sort_keys=True, indent=4),
            self.JSON_EXPECTED,
        )

    def test_config_addl_key(self):
        # This test creates a Configuration from an INI file with additional
        # keys that should get ignored in the instantiated object, but trigger
        # log warnings.

        ini_input = """
[instances]
agent

[manager]
instance = agent
port = 5000
role = manager
not_a_key = mhmmm
also_not_a_key = uh oh
"""
        ini_expected = """[instances]
agent

[manager]
instance = agent
role = MANAGER
port = 5000
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertTrue(config is not None)

        cfp = config.to_config_parser()
        with io.StringIO() as buf:
            cfp.write(buf)
            self.assertEqualStripped(buf.getvalue(), ini_expected)

        self.assertEqualStripped(
            self.logbuf.getvalue(),
            "warning: ignoring unexpected keys: also_not_a_key, not_a_key",
        )

    def test_config_ipv4_ipv6_instances(self):
        # This test creates a Configuration from an INI file with various IP addresses
        # and ports specified for the agents.

        ini_input = """[instances]
agent = 127.0.0.1:2151
agent2 = ::1:2151
agent3 = [::2]:2151

[manager]
instance = agent
role = MANAGER
port = 5000
"""
        ini_expected = """[instances]
agent = 127.0.0.1:2151
agent2 = ::1:2151
agent3 = ::2:2151

[manager]
instance = agent
role = MANAGER
port = 5000
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertTrue(config is not None)

        # Turning that back into a config parser should have expected content:
        cfp = config.to_config_parser()
        with io.StringIO() as buf:
            cfp.write(buf)
            self.assertEqualStripped(buf.getvalue(), ini_expected)

    def test_config_invalid_ipv4_instance(self):
        # This test creates a Configuration with an invalid IPv4 address

        ini_input = """[instances]
agent = 127.0.0.1.1:2151

[manager]
instance = agent
role = MANAGER
port = 5000
"""
        cfp = self.parser_from_string(ini_input)
        with self.assertRaisesRegex(
            ValueError,
            "'127.0.0.1.1' does not appear to be an IPv4 or IPv6 address",
        ):
            zeekclient.types.Configuration.from_config_parser(cfp)

    def test_config_invalid_ipv6_instance(self):
        # This test creates a Configuration with an invalid IPv6 address

        ini_input = """[instances]
agent = ::2/128:2151

[manager]
instance = agent
role = MANAGER
port = 5000
"""
        cfp = self.parser_from_string(ini_input)
        with self.assertRaisesRegex(
            ValueError,
            "'::2/128' does not appear to be an IPv4 or IPv6 address",
        ):
            zeekclient.types.Configuration.from_config_parser(cfp)

    def test_config_invalid_instances(self):
        ini_input = """
[instances]
agent = foo:

[manager]
instance = agent
port = 80
role = manager
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertFalse(config)

        self.assertEqualStripped(
            self.logbuf.getvalue(),
            'error: invalid spec for instance "agent": "foo:" should be <host>:<port>',
        )

    def test_config_missing_instance(self):
        ini_input = """
[instances]
agent

[manager]
role = manager
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertFalse(config)

        self.assertEqualStripped(
            self.logbuf.getvalue(),
            "error: omit instances section when skipping instances in node definitions",
        )

    def test_config_mixed_instances(self):
        ini_input = """
[manager]
role = manager

[worker]
role = worker
instance = agent1
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertFalse(config)

        self.assertEqualStripped(
            self.logbuf.getvalue(),
            "error: either all or no nodes must state instances",
        )

    def test_config_missing_role(self):
        ini_input = """
[instances]
agent

[manager]
instance = agent
port = 80
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertFalse(config)

        self.assertEqualStripped(
            self.logbuf.getvalue(),
            'error: invalid node "manager" configuration: node requires a role',
        )

    def test_config_invalid_role(self):
        ini_input = """
[instances]
agent

[manager]
instance = agent
port = 80
role = superintendent
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertFalse(config)

        self.assertEqualStripped(
            self.logbuf.getvalue(),
            'error: invalid node "manager" configuration: role "superintendent" is invalid',
        )

    def test_config_invalid_port_string(self):
        ini_input = """
[instances]
agent

[manager]
instance = agent
port = eighty
role = manager
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertFalse(config)

        self.assertEqualStripped(
            self.logbuf.getvalue(),
            'error: invalid node "manager" configuration: '
            'cannot convert "manager.port" value "eighty" to int',
        )

    def test_config_invalid_port_number(self):
        ini_input = """
[instances]
agent

[manager]
instance = agent
port = 70000
role = manager
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertFalse(config)

        self.assertEqualStripped(
            self.logbuf.getvalue(),
            'error: invalid node "manager" configuration: port 70000 outside valid range',
        )

    @patch("zeekclient.types.socket.gethostname", new=MagicMock(return_value="testbox"))
    def test_config_no_instances(self):
        ini_input = """
[manager]
role = manager
"""
        ini_expected = """
[instances]
agent-testbox

[manager]
instance = agent-testbox
role = MANAGER
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertTrue(config is not None)

        cfp = config.to_config_parser()
        with io.StringIO() as buf:
            cfp.write(buf)
            self.assertEqualStripped(buf.getvalue(), ini_expected)

    def test_config_missing_instance_section(self):
        ini_input = """
[manager]
instance = agent
role = manager

[logger]
instance = agent2
role = logger

[worker]
instance = agent
role = worker
"""
        ini_expected = """
[instances]
agent
agent2

[logger]
instance = agent2
role = LOGGER

[manager]
instance = agent
role = MANAGER

[worker]
instance = agent
role = WORKER
"""
        cfp = self.parser_from_string(ini_input)
        config = zeekclient.types.Configuration.from_config_parser(cfp)
        self.assertTrue(config is not None)

        cfp = config.to_config_parser()
        with io.StringIO() as buf:
            cfp.write(buf)
            self.assertEqualStripped(buf.getvalue(), ini_expected)



"""This verifies zeekclient's ability to load configurations and update
individual settings via command-line arguments, environment variables, and
files.
"""

import os
import tempfile
import unittest

import zeekclient


class TestConfig(unittest.TestCase):
    def setUp(self):
        self.config = zeekclient.config.Config()

    def test_basics(self):
        # One of each type:
        self.assertEqual(self.config.getint("client", "request_timeout_secs"), 20)
        self.assertEqual(
            self.config.getfloat("client", "peering_retry_delay_secs"),
            1.0,
        )
        self.assertEqual(self.config.getboolean("client", "rich_logging_format"), False)

    def test_update_from_file(self):
        with tempfile.NamedTemporaryFile("w", delete=False) as hdl:
            hdl.write("[client]\nrequest_timeout_secs = 10\n")
            hdl.close()
            self.config.update_from_file(hdl.name)
            self.assertEqual(self.config.getint("client", "request_timeout_secs"), 10)

    @unittest.mock.patch.dict(
        os.environ,
        {
            "ZEEK_CLIENT_CONFIG_SETTINGS": 'client.request_timeout_secs=23 server.FOO="1 2 3"',
        },
    )
    def test_update_from_env(self):
        self.config.update_from_env()
        self.assertEqual(self.config.getint("client", "request_timeout_secs"), 23)
        self.assertEqual(self.config.get("server", "FOO"), "1 2 3")

    def test_update_from_args(self):
        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(
            ["--set", "client.request_timeout_secs=42", "--set", "server.FOO=1 2 3"],
        )
        self.config.update_from_args(args)
        self.assertEqual(self.config.getint("client", "request_timeout_secs"), 42)
        self.assertEqual(self.config.get("server", "FOO"), "1 2 3")

    def test_update_from_args_controller_host(self):
        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "foo"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "foo")
        self.assertEqual(self.config.getint("controller", "port"), 2149)

        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "foo:"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "foo")
        self.assertEqual(self.config.getint("controller", "port"), 2149)

        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "127.0.0.1"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "127.0.0.1")
        self.assertEqual(self.config.getint("controller", "port"), 2149)

        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "127.0.0.1:"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "127.0.0.1")
        self.assertEqual(self.config.getint("controller", "port"), 2149)

        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "[fe80::1]"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "[fe80::1]")
        self.assertEqual(self.config.getint("controller", "port"), 2149)

        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "[fe80::1]:"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "[fe80::1]")
        self.assertEqual(self.config.getint("controller", "port"), 2149)

    def test_update_from_args_controller_port(self):
        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", ":2222"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "127.0.0.1")
        self.assertEqual(self.config.getint("controller", "port"), 2222)

    def test_update_from_args_controller_hostport(self):
        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "foo:2222"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "foo")
        self.assertEqual(self.config.getint("controller", "port"), 2222)

        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "127.0.0.1:2222"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "127.0.0.1")
        self.assertEqual(self.config.getint("controller", "port"), 2222)

        parser = zeekclient.cli.create_parser()
        args = parser.parse_args(["--controller", "[fe80::1]:2222"])
        self.config.update_from_args(args)
        self.assertEqual(self.config.get("controller", "host"), "[fe80::1]")
        self.assertEqual(self.config.getint("controller", "port"), 2222)



"""This verifies the Python-level representations of Zeek records in types module."""

import unittest

import zeekclient.brokertypes as bt
from zeekclient.types import (
    ClusterRole,
    Configuration,
    Instance,
    ManagementRole,
    Node,
    NodeOutputs,
    NodeStatus,
    Option,
    Result,
    State,
)


class TestTypes(unittest.TestCase):
    def assertHash(self, val):  # noqa: N802
        d = {val: 1}
        self.assertEqual(d[val], 1)

    def brokertype_roundtrip(self, data):
        return type(data).from_brokertype(data.to_brokertype())

    def test_cluster_role(self):
        val0 = ClusterRole.LOGGER
        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)

        self.assertTrue(ClusterRole.LOGGER < ClusterRole.MANAGER)
        with self.assertRaises(TypeError):
            _ = ClusterRole.LOGGER < State.PENDING

        val0_bt = val0.to_brokertype()
        self.assertEqual(val0_bt.to_py(), val0.qualified_name())

        self.assertHash(val0)

        with self.assertRaises(TypeError):
            ClusterRole.from_brokertype(bt.String("Oopsie::WORKER"))
        with self.assertRaises(TypeError):
            ClusterRole.from_brokertype(bt.String("Supervisor::OOPSIE"))

    def test_management_role(self):
        val0 = ManagementRole.AGENT
        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)

        val0_bt = val0.to_brokertype()
        self.assertEqual(val0_bt.to_py(), val0.qualified_name())

        self.assertHash(val0)

    def test_state(self):
        val0 = State.FAILED
        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)

        val0_bt = val0.to_brokertype()
        self.assertEqual(val0_bt.to_py(), val0.qualified_name())

        self.assertHash(val0)

    def test_option(self):
        val0 = Option("foo", "bar")
        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)

        self.assertHash(val0)

    def test_instance(self):
        val0 = Instance("instance1")
        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)
        self.assertEqual(val0.to_json_data(), {"name": "instance1"})

        val0 = Instance("instance1", "10.0.0.1")
        self.assertEqual(val0.to_json_data(), {"name": "instance1", "host": "10.0.0.1"})

        val0 = Instance("instance1", "10.0.0.1", 1234)
        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)
        self.assertEqual(
            val0.to_json_data(),
            {"name": "instance1", "host": "10.0.0.1", "port": 1234},
        )

        self.assertHash(val0)

        with self.assertRaisesRegex(TypeError, "unexpected Broker data"):
            Instance.from_brokertype(bt.String("not a vector..."))

    def test_node(self):
        val0 = Node("worker1", "instance1", ClusterRole.WORKER)
        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)

        val0 = Node(
            "worker1",
            "instance1",
            ClusterRole.WORKER,
            State.PENDING,
            port=1234,
            scripts=["foo/bar/baz.zeek"],
            options=[Option("foo", "bar")],
            interface="eth0",
            cpu_affinity=13,
            env={"FOO": "BAR"},
            metrics_port=9000,
        )
        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)

        self.assertHash(val0)

        with self.assertRaisesRegex(TypeError, "unexpected Broker data"):
            Node.from_brokertype(bt.String("not a vector..."))

    def test_configuration(self):
        val0 = Configuration()
        val0.instances.append(Instance("instance1"))
        val0.nodes.append(Node("worker1", "instance1", ClusterRole.WORKER))

        val1 = self.brokertype_roundtrip(val0)
        self.assertEqual(val0, val1)

        self.assertHash(val0)

    def test_node_status(self):
        # NodeStatus is not a SendableZeekType, so we make up the Brokertype
        # representation and only test instantiation from it. Same for the next
        # few types below.
        val0 = NodeStatus(
            "worker1",
            State.RUNNING,
            ManagementRole.NONE,
            ClusterRole.WORKER,
            pid=123,
            port=1234,
        )

        val1 = NodeStatus.from_brokertype(
            bt.Vector(
                [
                    bt.String("worker1"),
                    bt.Enum("Management::RUNNING"),
                    bt.Enum("Management::NONE"),
                    bt.Enum("Supervisor::WORKER"),
                    bt.Integer(123),
                    bt.Port(1234),
                ],
            ),
        )

        self.assertEqual(val0, val1)
        self.assertHash(val0)

    def test_result(self):
        val0 = Result(
            "reqid-1234",
            success=False,
            instance="instance1",
            data=bt.String("data"),
            error="an error",
            node="worker1",
        )

        val1 = Result.from_brokertype(
            bt.Vector(
                [
                    bt.String("reqid-1234"),
                    bt.Boolean(False),
                    bt.String("instance1"),
                    bt.String("data"),
                    bt.String("an error"),
                    bt.String("worker1"),
                ],
            ),
        )

        self.assertEqual(val0, val1)

    def test_node_outputs(self):
        val0 = NodeOutputs("stdout content", "stderr content")

        val1 = NodeOutputs.from_brokertype(
            bt.Vector(
                [
                    bt.String("stdout content"),
                    bt.String("stderr content"),
                ],
            ),
        )

        self.assertEqual(val0, val1)



"""Python-level representations of the records in policy/framework/management/types.zeek."""

import configparser
import enum
import shlex
import socket
from ipaddress import ip_address

from . import brokertypes as bt
from .logs import LOG
from .utils import make_uuid


class ConfigParserMixin:
    """Methods to create and render the object via ConfigParser instances."""

    @classmethod
    def from_config_parser(cls, cfp, section=None):
        """Instantiates an object of this class based on the given
        ConfigParser, and optional section name in it, as applicable.

        Raises ValueError if the provided configuration is invalid for the class
        to instantiate.
        """
        return None  # pragma: no cover

    def to_config_parser(self, cfp=None):
        """Returns this object in a ConfigParser instance. When the optional cfp
        argument is not None, the caller requests the implementation to add to
        the given parser, not create a new one.
        """
        return None  # pragma: no cover

    @staticmethod
    def _get(cfp, typ, section, *keys):
        """Typed config key/val retrieval, with support for key name aliases."""
        for key in keys:
            val = cfp.get(section, key, fallback=None)
            if val is not None:
                try:
                    return typ(val)
                except ValueError as err:
                    raise ValueError(
                        f'cannot convert "{section}.{key}" value "{val}" to {typ.__name__}',
                    ) from err
        return None


class SerializableZeekType:
    """An interface that supports serializing to and from Broker's data model.

    Objects of any class implementing this interface can be rendered to the
    Python-level Broker data model in the brokertypes module, and instantiated
    from it.
    """

    # We are not using abc.abstractmethod and friends here because the metaclass
    # magic they introduces clashes with multiple inheritance from other types,
    # affecting e.g. Enums below.
    def to_brokertype(self):
        """Returns a brokertype instance representing this object."""
        return None  # pragma: no cover

    @classmethod
    def from_brokertype(cls, data):
        """Returns an instance of this class for the given brokertype data.

        data: a brokertype instance

        Raises TypeError when the given data doesn't match the expected type.
        """
        return None  # pragma: no cover


class JsonableZeekType:
    """An interface for objects that can render themselves to JSON.

    This is not to be confused with the Broker-internal JSON representation for
    WebSockets. Instead, it refers to the JSON-formatted outputs zeek-client
    reports to the user.
    """

    def to_json_data(self):
        """Returns JSON-suitable datastructure representing the object."""
        return self.__dict__  # pragma: no cover


class ZeekType(SerializableZeekType, JsonableZeekType):
    """A does-it-all Zeek type."""


class Enum(ZeekType, enum.Enum):
    """A base class for Zeek's enums, with Python's enum features.

    This distinguishes the "flat" Python enums ("FOO") from the fully qualified
    way they're rendered via Zeek ("Some::Module::FOO"). To enable a Python enum
    to present the full qualification when sending into Broker, derivations
    reimplement the module_scope() class method to prefix with a scope string.
    """

    def __lt__(self, other):
        if self.__class__ != other.__class__:
            return NotImplemented
        return self.qualified_name() < other.qualified_name()

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.qualified_name() == other.qualified_name()
        )

    def __hash__(self):
        return hash((self.qualified_name(), self.value))

    def to_brokertype(self):
        scope = self.module_scope()
        scope = scope + "::" if scope else ""
        return bt.Enum(scope + self.name)

    def to_json_data(self):
        # A similar concern as above applies here, but the exact enum type will
        # often be clear from context and so the un-scoped name alone may
        # suffice.
        return self.name

    def qualified_name(self):
        scope = self.module_scope()
        scope = scope + "::" if scope else ""
        return scope + self.name

    @classmethod
    def lookup(cls, name):
        """Robust name-based lookup of an enum value.

        This removes any Zeek-land or Python-land qualifications, and
        automatically upper-cases the looked-up name.

        Raises KeyError if the requested enum value isn't defined.
        """
        name = name.split("::")[-1]
        name = name.split(".")[-1]
        return cls[name.upper()]

    @classmethod
    def module_scope(cls) -> str:  # pragma: no cover
        # Reimplement this in derived classes to convey the Zeek-level enum
        # scope. For example, for a Foo.BAR (or Foo::BAR, in Zeek) enum value,
        # this should return the string "Foo".
        raise AssertionError("reimplement module_scope() in your Enum derivative")

    @classmethod
    def from_brokertype(cls, data):
        # The argument should be a brokertype.Enum a scoped value such as
        # "Foo::VALUE".
        try:
            module = data.to_py().split("::", 1)[0]
            if module != cls.module_scope():
                raise ValueError(
                    f"module scope mismatch for {cls.__name__}: {module} != {cls.module_scope()}.",
                )
            return cls.lookup(data.to_py())
        except (ValueError, KeyError) as err:
            raise TypeError(
                f"unexpected enum value for {cls.__name__}: {repr(data)}",
            ) from err


class ClusterRole(Enum):
    """Equivalent of Supervisor::ClusterRole enum in Zeek"""

    NONE = 0
    LOGGER = 1
    MANAGER = 2
    PROXY = 3
    WORKER = 4

    @classmethod
    def module_scope(cls):
        return "Supervisor"


class ManagementRole(Enum):
    """Equivalent of Management::Role enum in Zeek"""

    NONE = 0
    AGENT = 1
    CONTROLLER = 2
    NODE = 3

    @classmethod
    def module_scope(cls):
        return "Management"


class State(Enum):
    """Equivalent of Management::State enum in Zeek"""

    PENDING = 0
    RUNNING = 1
    STOPPED = 2
    FAILED = 3
    CRASHED = 4
    UNKNOWN = 5

    @classmethod
    def module_scope(cls):
        return "Management"


class Option(ZeekType):
    """Equivalent of Management::Option."""

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.name == other.name
            and self.value == other.value
        )

    def __hash__(self):
        return hash((self.name, self.value))

    def to_brokertype(self):
        return bt.Vector([bt.String(self.name), bt.String(self.value)])

    @classmethod
    def from_brokertype(cls, data):
        return Option(*data.to_py())


class Instance(ZeekType):
    """Equivalent of Management::Instance."""

    def __init__(self, name, addr=None, port=None):
        self.name = name
        # This is a workaround until we've resolved addresses in instances
        self.host = "0.0.0.0"  # XXX needs proper optionality

        # If addr isn't a valid address, the ipaddress module will raise a ValueError
        if addr is not None and ip_address(addr):
            self.host = str(addr)
        self.port = port  # None or integer value; we always mean TCP

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.name == other.name
            and self.host == other.host
            and self.port == other.port
        )

    def __hash__(self):
        return hash((self.name, self.host, self.port))

    def to_brokertype(self):
        return bt.Vector(
            [
                bt.String(self.name),
                bt.Address(self.host),
                bt.from_py(self.port, typ=bt.Port),
            ],
        )

    def to_json_data(self):
        if self.port is not None:
            return self.__dict__

        # Here too, work around 0.0.0.0 until resolved
        if str(self.host) != "0.0.0.0":
            return {"name": self.name, "host": self.host}

        return {"name": self.name}

    @classmethod
    def from_brokertype(cls, data):
        try:
            name, addr, port = data.to_py()
            return Instance(name, addr, None if port is None else port.number)
        except ValueError as err:
            raise TypeError(
                f"unexpected Broker data for Instance object ({data})",
            ) from err


class Node(ZeekType, ConfigParserMixin):
    """Equivalent of Management::Node."""

    def __init__(
        self,
        name,
        instance,
        role,
        state=State.RUNNING,
        port=None,
        scripts=None,
        options=None,
        interface=None,
        cpu_affinity=None,
        env=None,
        metrics_port=None,
    ):
        self.name = name
        self.instance = instance
        self.role = role
        self.state = state
        self.port = port
        self.scripts = scripts
        self.options = options  # We use a list, Zeek record uses set
        self.interface = interface
        self.cpu_affinity = cpu_affinity
        self.env = env or {}
        self.metrics_port = metrics_port

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.name == other.name
            and self.instance == other.instance
            and self.role == other.role
            and self.state == other.state
            and self.port == other.port
            and self.scripts == other.scripts
            and self.options == other.options
            and self.interface == other.interface
            and self.cpu_affinity == other.cpu_affinity
            and self.env == other.env
            and self.metrics_port == other.metrics_port
        )

    def __hash__(self):
        scripts = tuple(self.scripts) if self.scripts else None
        options = tuple(self.options) if self.options else None
        env = None

        if self.env:
            env = ((key, self.env[key]) for key in sorted(self.env))

        return hash(
            (
                self.name,
                self.instance,
                self.role,
                self.state,
                self.port,
                scripts,
                options,
                self.interface,
                self.cpu_affinity,
                env,
                self.metrics_port,
            )
        )

    def to_brokertype(self):
        options = bt.NoneType()
        if self.options:
            options = bt.Set({opt.to_brokertype() for opt in self.options})

        return bt.Vector(
            [
                bt.String(self.name),
                bt.String(self.instance),
                self.role.to_brokertype(),
                self.state.to_brokertype(),
                bt.from_py(self.port, typ=bt.Port),
                bt.from_py(self.scripts),
                options,
                bt.from_py(self.interface),
                bt.from_py(self.cpu_affinity, typ=bt.Integer),
                bt.from_py(self.env),
                bt.from_py(self.metrics_port, typ=bt.Port),
            ]
        )

    def to_json_data(self):
        return {
            "name": self.name,
            "instance": self.instance,
            "role": self.role.to_json_data(),
            # We currently omit the state field since it has no effect on
            # cluster node operation.
            # 'state': self.state.to_json_data(),
            "port": self.port,
            "scripts": self.scripts,
            "options": self.options,
            "interface": self.interface,
            "cpu_affinity": self.cpu_affinity,
            "env": self.env,
            "metrics_port": self.metrics_port,
        }

    @classmethod
    def from_brokertype(cls, data):
        try:
            options = None
            if isinstance(data[6], bt.Set):
                options = [Option.from_brokertype(opt_data) for opt_data in data[6]]

            port = None
            if isinstance(data[4], bt.Port):
                port = data[4].number

            metrics_port = None
            if len(data) >= 11 and isinstance(data[10], bt.Port):
                metrics_port = data[10].number

            return Node(
                data[0].to_py(),  # name
                data[1].to_py(),  # instance
                ClusterRole.from_brokertype(data[2]),
                State.from_brokertype(data[3]),
                port,
                data[5].to_py(),  # scripts
                options,
                data[7].to_py(),  # interface
                data[8].to_py(),  # cpu_affinity
                data[9].to_py(),  # env
                metrics_port,
            )
        except (IndexError, TypeError, ValueError) as err:
            raise TypeError(f"unexpected Broker data for Node object ({data})") from err

    @classmethod
    def from_config_parser(cls, cfp, section=None):
        def get(typ, *keys):
            return cls._get(cfp, typ, section, *keys)

        name = section
        instance = get(str, "instance")
        role = get(str, "role", "type")

        # We currently ignore the node state, if provided. The Node class
        # defaults to 'RUNNING'.
        state = State.RUNNING
        if get(str, "state"):
            LOG.warning(
                'ignoring node "%s" state "%s" in configuration',
                name,
                get(str, "state"),
            )

        port = get(int, "port")
        scripts = None

        # The Node record type on the Zeek side features a set[Options] that we
        # don't use (yet).

        interface = get(str, "interface")
        cpu_affinity = get(int, "cpu_affinity")
        env = None
        metrics_port = get(int, "metrics_port")

        # Validate the specified values
        if not instance:
            # When a node features no instance name, default to
            # "agent-<hostname>", assuming the config targets host-local
            # deployment.
            hostname = socket.gethostname() or "localhost"
            instance = "agent-" + hostname

        if not role:
            raise ValueError("node requires a role")

        try:
            role = ClusterRole.lookup(role)
        except (AttributeError, KeyError) as err:
            raise ValueError(f'role "{role}" is invalid') from err

        # Optional values follow:

        # Ports are optional and filled in by the controller, assuming
        # Management::Controller::auto_assign_ports is enabled. But when
        # present, we validate:
        if port is not None and (port < 1 or port > 65535):
            raise ValueError(f"port {port} outside valid range")

        # Same for metrics ports.
        if metrics_port is not None and (metrics_port < 1 or metrics_port > 65535):
            raise ValueError(f"metrics port {port} outside valid range")

        try:
            # We support multiple scripts as a simple space-separated sequence
            # of filenames, with possible quotation marks for strings with
            # spaces. The shlex module provides a convenient way to parse these.
            val = get(str, "scripts")
            if val:
                scripts = sorted(shlex.split(val))
        except (AttributeError, KeyError) as err:
            raise ValueError(f'scripts value "{val}" is invalid') from err

        try:
            # An environment variable dictionary is represented as a single
            # config value: a space-separated sequence of <var>=<val> strings,
            # possibly with quotation marks around the val. shlex helps here
            # too: shlex.split('foo=bar=baz blum="foo bar baz"') yields
            # ['foo=bar=baz', 'blum=foo bar baz']
            val = get(str, "env")
            if val:
                env = {}
                for item in shlex.split(val):
                    key, kval = item.split("=", 1)
                    env[key] = kval
        except (AttributeError, KeyError, ValueError) as err:
            raise ValueError(f'env value "{val}" is invalid') from err

        # Warn about unexpected keys:
        cfp_subset = cfp[section] if section else cfp
        keys = set(cfp_subset.keys())
        keys -= {
            "instance",
            "role",
            "scripts",
            "port",
            "interface",
            "cpu_affinity",
            "env",
            "metrics_port",
        }

        if len(keys) > 0:
            LOG.warning("ignoring unexpected keys: %s", ", ".join(sorted(keys)))

        return Node(
            name=name,
            instance=instance,
            role=role,
            state=state,
            port=port,
            scripts=scripts,
            interface=interface,
            cpu_affinity=cpu_affinity,
            env=env,
            metrics_port=metrics_port,
        )

    def to_config_parser(self, cfp=None):
        if cfp is None:
            cfp = configparser.ConfigParser(allow_no_value=True)

        if self.name in cfp.sections():
            cfp.remove_section(self.name)

        cfp.add_section(self.name)

        cfp.set(self.name, "instance", self.instance)
        cfp.set(self.name, "role", self.role.name)

        # Skip state for the moment, it has no operational effect
        # if self.state is not None:
        #    cfp.set(self.name, 'state', self.state.name)

        if self.port is not None:
            cfp.set(self.name, "port", str(self.port))

        if self.scripts:
            # See if any of the script paths contain spaces, and use quotation
            # marks if so. This does not escape quotation marks or deal with
            # other "difficult" characters.
            scripts = []

            for script in sorted(self.scripts):
                if len(script.split()) > 1:
                    script = '"' + script + '"'
                scripts.append(script)

            cfp.set(self.name, "scripts", " ".join(scripts))

        if self.interface is not None:
            cfp.set(self.name, "interface", self.interface)

        if self.cpu_affinity is not None:
            cfp.set(self.name, "cpu_affinity", str(self.cpu_affinity))

        if self.env:
            # If the value has whitespace, use key="val". As with scripts above,
            # this does not deal with more complicated escaping/characters.
            env = []

            for key in sorted(self.env.keys()):
                val = self.env[key]
                if len(str(val).split()) > 1:
                    val = '"' + val + '"'

                env.append(f"{key}={val}")

            cfp.set(self.name, "env", " ".join(env))

        if self.metrics_port is not None:
            cfp.set(self.name, "metrics_port", str(self.metrics_port))

        return cfp


class Configuration(ZeekType, ConfigParserMixin):
    """Equivalent of Management::Configuration."""

    def __init__(self):
        self.id = make_uuid()
        # The following are sets in the Zeek record equivalents. We could
        # reflect this, but handling lists is easier. They do get serialized
        # to/from sets.
        self.instances = []
        self.nodes = []

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.id == other.id
            and self.instances == other.instances
            and self.nodes == other.nodes
        )

    def __hash__(self):
        return hash((self.id, tuple(self.instances), tuple(self.nodes)))

    def to_brokertype(self):
        return bt.Vector(
            [
                bt.String(self.id),
                bt.Set({inst.to_brokertype() for inst in self.instances}),
                bt.Set({node.to_brokertype() for node in self.nodes}),
            ],
        )

    def to_json_data(self):
        return {
            "id": self.id,
            "instances": [inst.to_json_data() for inst in sorted(self.instances)],
            "nodes": [node.to_json_data() for node in sorted(self.nodes)],
        }

    @classmethod
    def from_brokertype(cls, data):
        res = Configuration()
        res.id = data[0].to_py()
        for inst_data in data[1]:
            res.instances.append(Instance.from_brokertype(inst_data))
        for node_data in data[2]:
            res.nodes.append(Node.from_brokertype(node_data))
        res.instances.sort()
        res.nodes.sort()
        return res

    @classmethod
    def from_config_parser(cls, cfp, _section=None):
        config = Configuration()

        # The nodes in this configuration that do not specify an instance.
        # This is a convenience this client offers, so let's be consistent:
        # if we use this feature, the entire config must be instance-free.
        instance_free_nodes = set()

        for section in cfp.sections():
            if section == "instances":
                # The [instances] section is special: each key in it is the name of
                # an instance, each val is the host:port pair where its agent is
                # listening. The val may be absent when it's an instance that
                # connects to the controller.
                for key, val in cfp.items("instances"):
                    if not val:
                        config.instances.append(Instance(key))
                    else:
                        hostport = val
                        host, _, port = hostport.rpartition(":")
                        if host == "" or port == "":
                            LOG.error(
                                'invalid spec for instance "%s": "%s" should be <host>:<port>',
                                key,
                                val,
                            )
                            return None
                        # remove brackets to support [ipv6]:port formats
                        host = host.strip("[] ")
                        port = port.strip()
                        config.instances.append(Instance(key, host, port))
                continue

            # All keys for sections other than "instances" need to have a value.
            for key, val in cfp.items(section):
                if val is None:
                    LOG.error("config item %s.%s needs a value", section, key)
                    return None

            # The other sections are cluster nodes. Each section name corresponds to
            # a node name, with the keys being one of "type", "instance", etc.
            if section in [node.name for node in config.nodes]:
                LOG.warning(
                    'node "%s" defined more than once, skipping repeats"',
                    section,
                )
                continue

            try:
                if "instance" not in cfp[section]:
                    instance_free_nodes.add(section)
                config.nodes.append(Node.from_config_parser(cfp, section))
            except ValueError as err:
                LOG.error('invalid node "%s" configuration: %s', section, err)
                return None

        # Reject if this config mixes instance-free and instance-claiming nodes,
        # or if it uses an instances section while omitting instances in nodes.
        if len(instance_free_nodes) > 0:
            if len(instance_free_nodes) != len(config.nodes):
                LOG.error("either all or no nodes must state instances")
                return None
            if "instances" in cfp.sections():
                LOG.error(
                    "omit instances section when skipping instances in node definitions",
                )
                return None

        # When the configuration has no "instances" section, then any instance
        # names given in node sections imply corresponding instances whose
        # agents connect to the controller. That is, the instances section is
        # just a redundant listing of the instance names and we can synthesize
        # it:
        if "instances" not in cfp.sections():
            names = set()
            for node in config.nodes:
                names.add(node.instance)
            config.instances = sorted([Instance(name) for name in names])

        # We don't cross-check the set of instances claimed by the nodes vs the
        # set of instances declared in the config, because the controller
        # already does this.

        return config

    def to_config_parser(self, cfp=None):
        if cfp is None:
            cfp = configparser.ConfigParser(allow_no_value=True)

        if "instances" in cfp.sections():
            cfp.remove_section("instances")

        if self.instances:
            cfp.add_section("instances")
            for inst in sorted(self.instances):
                if inst.port is not None:
                    # An instance the controller connects to
                    cfp.set("instances", inst.name, f"{inst.host}:{inst.port}")
                else:
                    # An instance connecting to the controller
                    cfp.set("instances", inst.name)

        for node in sorted(self.nodes):
            node.to_config_parser(cfp)

        return cfp


class NodeStatus(SerializableZeekType):
    """Equivalent of Management::NodeState."""

    def __init__(
        self,
        node,
        state,
        mgmt_role,
        cluster_role,
        pid=None,
        port=None,
        metrics_port=None,
    ):
        self.node = node  # A string containing the name of the node
        self.state = state  # A State enum value
        self.mgmt_role = mgmt_role  # A ManagementRole enum value
        self.cluster_role = cluster_role  # A ClusterRole enum value
        self.pid = pid  # A numeric process ID
        self.port = port  # A numeric (TCP) port
        self.metrics_port = metrics_port  # A numeric (TCP) port for Prometheus

    def __lt__(self, other):
        return self.node < other.node

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.node == other.node
            and self.state == other.state
            and self.mgmt_role == other.mgmt_role
            and self.cluster_role == other.cluster_role
            and self.pid == other.pid
            and self.port == other.port
            and self.metrics_port == other.metrics_port
        )

    def __hash__(self):
        return hash(
            (
                self.node,
                self.state,
                self.mgmt_role,
                self.cluster_role,
                self.pid,
                self.port,
                self.metrics_port,
            )
        )

    def to_brokertype(self):
        # In normal operation we only ever receive NodeStates, but for testing
        # it helps to be able to serialize.
        pid = bt.NoneType() if self.pid is None else bt.Integer(self.pid)
        port = bt.NoneType() if self.port is None else bt.Port(self.port)
        metrics_port = (
            bt.NoneType() if self.metrics_port is None else bt.Port(self.metrics_port)
        )

        return bt.Vector(
            [
                bt.String(self.node),
                self.state.to_brokertype(),
                self.mgmt_role.to_brokertype(),
                self.cluster_role.to_brokertype(),
                pid,
                port,
                metrics_port,
            ]
        )

    @classmethod
    def from_brokertype(cls, data):
        port = data[5].to_py()
        if port is not None:
            port = port.number

        metrics_port = None
        if len(data) >= 7:
            metrics_port = data[6].to_py()
            if metrics_port is not None:
                metrics_port = metrics_port.number

        return NodeStatus(
            data[0].to_py(),
            State.from_brokertype(data[1]),
            ManagementRole.from_brokertype(data[2]),
            ClusterRole.from_brokertype(data[3]),
            data[4].to_py(),
            port,
            metrics_port,
        )


class Result(SerializableZeekType):
    """Equivalent of Management::Result."""

    def __init__(
        self,
        reqid,
        success=True,
        instance=None,
        data=None,
        error=None,
        node=None,
    ):
        self.reqid = reqid
        self.success = success
        self.instance = instance
        self.data = data
        self.error = error
        self.node = node

    def __lt__(self, other):
        """Support sorting. Sort first by instance name the result comes from,
        second by the node name if present.
        """
        if self.instance is None and other.instance is not None:
            return False
        if self.instance is not None and other.instance is None:
            return True
        if self.instance is not None and other.instance is not None:
            if self.instance < other.instance:
                return True
            if self.instance > other.instance:
                return False

        # Be more specific if we have a node name -- we can use it to sort when
        # two results come from the same instance.
        if self.node is not None and other.node is not None:
            return self.node < other.node

        return False

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.reqid == other.reqid
            and self.success == other.success
            and self.instance == other.instance
            and self.data == other.data
            and self.error == other.error
            and self.node == other.node
        )

    def hash(self):
        return hash(
            (self.reqid, self.success, self.instance, self.data, self.error, self.node),
        )

    def to_brokertype(self):
        # In normal operation we only ever receive Results, but for testing it
        # helps to be able to serialize.
        instance = bt.NoneType() if self.instance is None else bt.String(self.instance)

        data = bt.NoneType()
        if self.data is not None:
            # This is any-typed in Zeek and so a bit special: it is up to the
            # caller what exactly this is, an it is assumed to already be in
            # Brokertype format. We just pass it through.
            data = self.data

        error = bt.NoneType() if self.error is None else bt.String(self.error)
        node = bt.NoneType() if self.node is None else bt.String(self.node)

        return bt.Vector(
            [
                bt.String(self.reqid),
                bt.Boolean(self.success),
                instance,
                data,
                error,
                node,
            ],
        )

    @classmethod
    def from_brokertype(cls, data):
        # The data field gets special treatment since it can be of any
        # type. When it's a brokertype.NoneType (i.e., not present), we turn it
        # into None, since that simplifies its handling. Otherwise we leave it
        # untouched: the correct type to deserialize into will become clear
        # later from surrounding context.
        res_data = data[3]

        if isinstance(res_data, bt.NoneType):
            res_data = None

        return Result(
            reqid=data[0].to_py(),
            success=data[1].to_py(),
            instance=data[2].to_py(),
            data=res_data,
            error=data[4].to_py(),
            node=data[5].to_py(),
        )


class NodeOutputs(SerializableZeekType):
    """Equivalent of Management::NodeOutputs."""

    def __init__(self, stdout, stderr):
        self.stdout = stdout
        self.stderr = stderr

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.stdout == other.stdout
            and self.stderr == other.stderr
        )

    def hash(self):
        return hash((self.stdout, self.stderr))

    def to_brokertype(self):
        # In normal operation we only ever receive NodeOutputs, but for testing
        # it helps to be able to serialize.
        return bt.Vector(
            [
                bt.String(self.stdout),
                bt.String(self.stderr),
            ],
        )

    @classmethod
    def from_brokertype(cls, data):
        return NodeOutputs(*(data.to_py()))



"""This verifies the behavior of the types provied by the brokertypes module."""

import datetime
import unittest

from zeekclient.brokertypes import (
    Address,
    Boolean,
    Count,
    DataMessage,
    Enum,
    ErrorMessage,
    HandshakeAckMessage,
    HandshakeMessage,
    Integer,
    NoneType,
    Port,
    Real,
    Set,
    String,
    Subnet,
    Table,
    Timespan,
    Timestamp,
    Vector,
    ZeekEvent,
    from_py,
    ipaddress,
    unserialize,
)


class TestBrokertypes(unittest.TestCase):
    def assertEqualRoundtrip(self, data):  # noqa: N802
        # This verifies for the given Brokertype object that it can serialize
        # into Broker's wire format, unserialize, and yield an identical object.
        output = type(data).unserialize(data.serialize())
        self.assertEqual(data, output)

    def assertHash(self, val):  # noqa: N802
        d = {val: 1}
        self.assertEqual(d[val], 1)

    def test_none(self):
        self.assertEqual(NoneType(None), NoneType())
        self.assertEqual(NoneType().to_py(), None)
        self.assertEqual(NoneType(), from_py(None))

        self.assertNotEqual(NoneType, None)

        self.assertEqualRoundtrip(NoneType())

        self.assertFalse(NoneType() < NoneType())
        self.assertHash(NoneType())

    def test_boolean(self):
        self.assertEqual(Boolean(True), Boolean(True))
        self.assertEqual(Boolean(True), Boolean("true"))
        self.assertEqual(Boolean(True).to_py(), True)
        self.assertEqual(Boolean(True), from_py(True))

        self.assertNotEqual(Boolean(True), Boolean(False))
        self.assertNotEqual(Boolean(True), True)

        self.assertEqualRoundtrip(Boolean(True))
        self.assertEqualRoundtrip(Boolean(False))

        self.assertTrue(Boolean(False) < Boolean(True))
        self.assertFalse(Boolean(False) > Boolean(True))
        self.assertHash(Boolean(True))

    def test_count(self):
        self.assertEqual(Count(10), Count(10))
        self.assertEqual(Count(10), Count("10"))
        self.assertEqual(Count(10).to_py(), 10)
        self.assertEqual(Count(10), from_py(10, Count))

        self.assertNotEqual(Count(10), Count(1))
        self.assertNotEqual(Count(10), 10)

        self.assertEqualRoundtrip(Count(10))

        for val in (-10, "hello"):
            with self.assertRaises(ValueError):
                Count(val)

        self.assertTrue(Count(1) < Count(10))
        self.assertTrue(Count(10) > Count(1))
        self.assertHash(Count(10))

    def test_integer(self):
        self.assertEqual(Integer(10), Integer(10))
        self.assertEqual(Integer(10), Integer("10"))
        self.assertEqual(Integer(10).to_py(), 10)
        self.assertEqual(Integer(10), from_py(10))

        self.assertNotEqual(Integer(10), Integer(1))
        self.assertNotEqual(Integer(10), 10)

        self.assertEqualRoundtrip(Integer(10))

        self.assertTrue(Integer(1) < Integer(10))
        self.assertTrue(Integer(10) > Integer(1))
        self.assertHash(Integer(10))

    def test_real(self):
        self.assertEqual(Real(10.1), Real(10.1))
        self.assertEqual(Real(10.1), Real("10.1"))
        self.assertEqual(Real(10.1).to_py(), 10.1)
        self.assertEqual(Real(10.1), from_py(10.1))

        self.assertNotEqual(Real(10.0), Real(10.1))
        self.assertNotEqual(Real(1.1), 1.1)

        self.assertEqualRoundtrip(Real(1.1))

        with self.assertRaises(ValueError):
            Real("hello")

        self.assertTrue(Real(1) < Real(10))
        self.assertTrue(Real(10) > Real(1))
        self.assertHash(Real(10.1))

    def test_timespan(self):
        self.assertEqual(Timespan("10s"), Timespan("10s"))
        self.assertEqual(Timespan("10s").to_py(), datetime.timedelta(seconds=10))

        self.assertEqual(
            Timespan("10.5d").to_py(),
            datetime.timedelta(days=10, hours=12),
        )
        self.assertEqual(
            Timespan(datetime.timedelta(microseconds=1)),
            Timespan("1000ns"),
        )
        self.assertEqual(Timespan(datetime.timedelta(milliseconds=1)), Timespan("1ms"))
        self.assertEqual(Timespan(datetime.timedelta(seconds=1)), Timespan("1000ms"))
        self.assertEqual(Timespan(datetime.timedelta(seconds=1)), Timespan("1s"))
        self.assertEqual(Timespan(datetime.timedelta(minutes=1)), Timespan("1min"))
        self.assertEqual(Timespan(datetime.timedelta(hours=1)), Timespan("1h"))
        self.assertEqual(Timespan(datetime.timedelta(days=1)), Timespan("1d"))
        self.assertEqual(Timespan(datetime.timedelta(weeks=1)), Timespan("7d"))

        self.assertNotEqual(Timespan("10s"), Timespan("20s"))
        self.assertNotEqual(Timespan("10s"), Timespan("10ms"))
        self.assertNotEqual(Timespan("10s"), 10)

        self.assertEqualRoundtrip(Timespan("10h"))

        for val in "oops", "10nano", "1a":
            with self.assertRaises(ValueError):
                Timespan(val)

        self.assertTrue(Timespan("1ns") < Timespan("1ms"))
        self.assertTrue(Timespan("1ms") < Timespan("1s"))
        self.assertTrue(Timespan("1s") < Timespan("1h"))
        self.assertTrue(Timespan("1h") < Timespan("1d"))

        self.assertHash(Timespan("10d"))

    def test_timestamp(self):
        ts1, ts2 = "2022-10-20T01:02:03.004", "2022-10-21T00:00:00.000"
        self.assertEqual(Timestamp(ts1), Timestamp(ts1))
        self.assertEqual(Timestamp(ts1).to_py(), datetime.datetime.fromisoformat(ts1))
        self.assertEqual(
            Timestamp(ts1),
            Timestamp(datetime.datetime.fromisoformat(ts1)),
        )

        self.assertNotEqual(Timestamp(ts1), Timestamp(ts2))
        self.assertNotEqual(Timestamp(ts1).to_py(), ts1)

        self.assertEqualRoundtrip(Timestamp(ts1))

        with self.assertRaises(ValueError):
            Timestamp("oops")

        self.assertTrue(Timestamp(ts1) < Timestamp(ts2))

    def test_string(self):
        self.assertEqual(String("10"), String("10"))
        self.assertEqual(String(True), String(True))
        self.assertEqual(String("23"), String(23))
        self.assertEqual(String("23").to_py(), "23")
        self.assertEqual(String("foo"), from_py("foo"))

        self.assertNotEqual(String("10"), String("20"))

        self.assertEqualRoundtrip(String("foo"))

        self.assertTrue(String("bar") < String("foo"))
        self.assertTrue(String("foo") > String("bar"))
        self.assertHash(String("foo"))

    def test_enum(self):
        self.assertEqual(Enum("Foo"), Enum("Foo"))
        self.assertEqual(Enum("Foo").to_py(), "Foo")

        self.assertNotEqual(Enum("Foo"), Enum("FOO"))
        self.assertNotEqual(Enum("Foo"), Enum("Bar"))

        self.assertEqualRoundtrip(Enum("Foo::bar"))

        self.assertTrue(Enum("FOO::bar") < Enum("FOO::baz"))
        self.assertTrue(Enum("FOO::baz") > Enum("FOO::bar"))
        self.assertHash(Enum("Foo::bar"))

    def test_address(self):
        self.assertEqual(Address("127.0.0.1"), Address("127.0.0.1"))
        self.assertEqual(
            Address(ipaddress.ip_address("127.0.0.1")),
            Address("127.0.0.1"),
        )
        self.assertEqual(
            Address(ipaddress.ip_address("2001:db8::")),
            Address("2001:db8::"),
        )
        self.assertEqual(
            Address("127.0.0.1").to_py(),
            ipaddress.ip_address("127.0.0.1"),
        )
        self.assertEqual(
            Address("127.0.0.1"),
            from_py(ipaddress.ip_address("127.0.0.1")),
        )

        self.assertNotEqual(Address("127.0.0.1"), Address("10.0.0.1"))

        self.assertEqualRoundtrip(Address("10.0.0.1"))

        for val in ("foo", "10.0.0.0/8"):
            with self.assertRaises(ValueError):
                Address(val)

        self.assertTrue(Address("1.0.0.1") < Address("1.0.0.2"))
        self.assertTrue(Address("1.0.0.2") > Address("1.0.0.1"))
        self.assertHash(Address("127.0.0.1"))

    def test_subnet(self):
        self.assertEqual(Subnet("10.0.0.0/8"), Subnet("10.0.0.0/8"))
        self.assertEqual(
            Subnet(ipaddress.ip_network("10.0.0.0/8")),
            Subnet("10.0.0.0/8"),
        )
        self.assertEqual(
            Subnet(ipaddress.ip_network("2001:db8::/32")),
            Subnet("2001:db8::/32"),
        )
        self.assertEqual(
            Subnet("10.0.0.0/8").to_py(),
            ipaddress.ip_network("10.0.0.0/8"),
        )
        self.assertEqual(
            Subnet("10.0.0.0/8"),
            from_py(ipaddress.ip_network("10.0.0.0/8")),
        )

        self.assertNotEqual(Subnet("10.0.0.0/8"), Subnet("10.0.0.1"))

        self.assertEqualRoundtrip(Subnet("10.0.0.0/8"))

        for val in ("foo", "10.0.0.0/64", "10.0.0.1/8"):
            with self.assertRaises(ValueError):
                Subnet(val)

        self.assertTrue(Subnet("10.0.0.0/8") < Subnet("10.0.0.0/16"))
        self.assertTrue(Subnet("10.0.0.0/16") > Subnet("10.0.0.0/8"))
        self.assertHash(Subnet("10.0.0.0/8"))

    def test_port(self):
        self.assertEqual(Port(10), Port(10))
        self.assertEqual(Port(10), Port("10"))
        self.assertEqual(Port(10, Port.Proto.TCP), Port(10, Port.Proto.TCP))
        self.assertEqual(Port(10, Port.Proto.UDP), Port(10, Port.Proto.UDP))
        self.assertEqual(Port(10, Port.Proto.ICMP), Port(10, Port.Proto.ICMP))
        self.assertEqual(Port(10, Port.Proto.UNKNOWN), Port(10, Port.Proto.UNKNOWN))
        self.assertEqual(Port(10).to_py(), Port(10))

        self.assertEqualRoundtrip(Port(443))
        self.assertEqualRoundtrip(Port(53, Port.Proto.UDP))

        self.assertNotEqual(Port(10), Port(20))
        self.assertNotEqual(Port(10), Port(10, Port.Proto.UDP))
        self.assertNotEqual(Port(10, Port.Proto.UDP), Port(10, Port.Proto.ICMP))
        self.assertNotEqual(Port(10, Port.Proto.UDP), Port(10, Port.Proto.UNKNOWN))

        with self.assertRaises(ValueError):
            Port("oops", Port.Proto.TCP)
        with self.assertRaises(ValueError):
            Port(70000)
        with self.assertRaises(TypeError):
            Port(10, "tcp")

        self.assertTrue(Port(10) < Port(20))
        self.assertTrue(Port(20) > Port(10))
        self.assertTrue(Port(20, Port.Proto.TCP) < Port(10, Port.Proto.UDP))
        self.assertTrue(Port(20, Port.Proto.UDP) < Port(10, Port.Proto.ICMP))
        self.assertHash(Port(10))

    def test_vector(self):
        val = Vector([from_py(1), from_py("foo"), from_py(True)])

        self.assertEqual(val, val)
        self.assertEqual(Vector([String("foo")]).to_py(), ["foo"])
        self.assertEqual(Vector([String("foo")]), from_py(["foo"]))

        self.assertNotEqual(Vector([from_py(1), from_py("foo")]), Vector([from_py(1)]))
        self.assertNotEqual(
            Vector([from_py(1), from_py("foo")]),
            Vector([from_py(1), from_py("noo")]),
        )

        self.assertEqualRoundtrip(Vector([from_py(1), from_py("foo"), from_py(True)]))

        for _ in val:
            pass
        self.assertEqual(len(val), 3)
        self.assertEqual(val[0], Integer(1))

        self.assertTrue(Vector([from_py(1)]) < Vector([from_py(2)]))
        self.assertTrue(Vector([from_py(1)]) < Vector([from_py(1), from_py("foo")]))
        self.assertHash(val)

        for val in (23, [23]):
            with self.assertRaises(TypeError):
                Vector(val)

    def test_set(self):
        val = Set({from_py(1), from_py("foo"), from_py(True)})

        self.assertEqual(val, val)
        self.assertEqual(Set({String("foo")}).to_py(), {"foo"})
        self.assertEqual(Set({String("foo")}), from_py({"foo"}))

        self.assertNotEqual(Set({from_py(1), from_py("foo")}), Set({from_py(1)}))
        self.assertNotEqual(
            Set({from_py(1), from_py("foo")}),
            Set({from_py(1), from_py("noo")}),
        )

        self.assertEqualRoundtrip(Set({from_py(1), from_py("foo"), from_py(True)}))

        for _ in val:
            pass
        self.assertEqual(len(val), 3)
        self.assertTrue(Integer(1) in val)

        self.assertTrue(Set({from_py(1)}) < Set({from_py(2)}))
        self.assertTrue(Set({from_py(1)}) < Set({from_py(1), from_py("foo")}))
        self.assertHash(val)

        for val in (23, {23}):
            with self.assertRaises(TypeError):
                Set(val)

    def test_table(self):
        val = Table({from_py("foo"): from_py(1), from_py("bar"): from_py(2)})

        self.assertEqual(val, val)
        self.assertEqual(
            Table({from_py("foo"): from_py(1), from_py("bar"): from_py(2)}).to_py(),
            {"foo": 1, "bar": 2},
        )
        self.assertEqual(Table({from_py("foo"): from_py(1)}), from_py({"foo": 1}))

        self.assertNotEqual(
            Table({from_py("foo"): from_py(1), from_py("bar"): from_py(2)}),
            Table({from_py("foo"): from_py(1), from_py("bar"): from_py(3)}),
        )
        self.assertNotEqual(
            Table({from_py("foo"): from_py(1), from_py("bar"): from_py(2)}),
            Table({from_py("foo"): from_py(1), from_py("baz"): from_py(2)}),
        )

        self.assertEqualRoundtrip(
            Table({from_py("foo"): from_py(1), from_py("bar"): from_py(2)}),
        )
        for _ in val:
            pass
        for _ in val.keys():
            pass
        for _ in val.values():
            pass
        for _, _ in val.items():
            pass
        self.assertEqual(len(val), 2)
        self.assertTrue(String("foo") in val)

        self.assertFalse(
            Table({from_py("foo"): from_py(1)}) < Table({from_py("foo"): from_py(1)}),
        )
        self.assertTrue(
            Table({from_py("foo"): from_py(1)}) < Table({from_py("foo"): from_py(2)}),
        )
        self.assertTrue(
            Table({from_py("bar"): from_py(1)})
            < Table({from_py("foo"): from_py(1), from_py("bar"): from_py(1)}),
        )
        self.assertTrue(
            Table({from_py("foo"): from_py(1)})
            < Table({from_py("foo"): from_py(1), from_py("bar"): from_py(2)}),
        )
        self.assertTrue(
            Table({from_py("aaa"): from_py(1)})
            < Table({from_py("foo"): from_py(1), from_py("bar"): from_py(2)}),
        )
        self.assertHash(val)

        for val in (23, {"foo": 23}):
            with self.assertRaises(TypeError):
                Table(val)

    def test_zeek_event(self):
        evt = ZeekEvent("Test::event", from_py("hello"), from_py(42), from_py(True))
        self.assertTrue(isinstance(evt, Vector))
        self.assertEqual(
            ZeekEvent("Test::event", from_py("hello"), from_py(42), from_py(True)),
            ZeekEvent("Test::event", from_py("hello"), from_py(42), from_py(True)),
        )
        self.assertNotEqual(
            ZeekEvent("Test::event", from_py("hello"), from_py(42), from_py(True)),
            ZeekEvent("Test::event2", from_py("hello"), from_py(42), from_py(True)),
        )
        self.assertNotEqual(
            ZeekEvent("Test::event", from_py("hello"), from_py(42), from_py(True)),
            ZeekEvent("Test::event", from_py("hello"), from_py(42)),
        )
        self.assertNotEqual(
            ZeekEvent("Test::event", from_py("hello"), from_py(43)),
            ZeekEvent("Test::event", from_py("hello"), from_py(42)),
        )

        self.assertEqualRoundtrip(evt)

        vec = Vector.unserialize(evt.serialize())
        evt2 = ZeekEvent.from_vector(vec)
        self.assertEqual(evt, evt2)

        with self.assertRaises(TypeError):
            ZeekEvent.from_vector(String("not a vector"))

        with self.assertRaises(TypeError):
            ZeekEvent("foo", 1)

    def test_zeek_event_from_vector_metadata(self):
        md_vec = Vector([Vector([from_py(12344242), from_py("truth")])])
        args_vec = Vector()
        ev_vec = Vector([from_py("Test::event"), args_vec, md_vec])
        vec = Vector([from_py(1), from_py(1), ev_vec])

        ev = ZeekEvent.from_vector(vec)
        self.assertEqual(ev.name, "Test::event")
        self.assertEqual(ev.args, [])

    def test_zeek_event_from_vector_invalid(self):
        test_cases = [
            ("missing args", Vector([from_py("Test::event")])),
            ("wrong name type", Vector([from_py(1), Vector()])),
            ("wrong args type", Vector([from_py("Test::event"), from_py("string")])),
        ]

        for name, ev_vec in test_cases:
            with self.subTest(msg=name):
                vec = Vector([from_py(1), from_py(1), ev_vec])
                with self.assertRaises(TypeError):
                    ZeekEvent.from_vector(vec)

    def test_handshake_message(self):
        self.assertEqual(
            HandshakeMessage(["foo", "bar"]),
            HandshakeMessage(["foo", String("bar")]),
        )
        self.assertNotEqual(HandshakeMessage(["foo", "bar"]), HandshakeMessage(["foo"]))
        self.assertEqualRoundtrip(HandshakeMessage(["foo", "bar"]))

    def test_handshake_ack_message(self):
        self.assertEqual(
            HandshakeAckMessage("aaaa", "1.0"),
            HandshakeAckMessage("aaaa", "1.0"),
        )
        self.assertNotEqual(
            HandshakeAckMessage("aaaa", "1.0"),
            HandshakeAckMessage("bbbb", "1.0"),
        )
        self.assertEqualRoundtrip(HandshakeAckMessage("aaaa", "1.0"))

    def test_data_message(self):
        self.assertEqual(
            DataMessage("foo", String("test")),
            DataMessage("foo", String("test")),
        )
        self.assertNotEqual(
            DataMessage("foo", String("test")),
            DataMessage("foo", String("other")),
        )
        self.assertEqualRoundtrip(DataMessage("foo", String("test")))

    def test_error_message(self):
        msg1 = ErrorMessage("deserialization_failed", "this is where you failed")
        msg2 = ErrorMessage("deserialization_failed", "this is where you also failed")
        self.assertEqual(msg1, msg1)
        self.assertNotEqual(msg1, msg2)
        self.assertEqualRoundtrip(msg1)

    def test_type_lt(self):
        # Any brokertyped data value can be compared to any other, but not to
        # unrelated types.
        self.assertTrue(Address("127.0.0.1") < Boolean(True))
        self.assertTrue(Boolean(True) < Count(1))
        self.assertTrue(Count(1) < Integer(1))
        self.assertTrue(Integer(1) < Real(1))
        self.assertTrue(Real(1) < Set())
        self.assertTrue(Set() < String("foo"))
        self.assertTrue(String("foo") < Subnet("10.0.0.0/8"))
        self.assertTrue(Subnet("10.0.0.0/8") < Timespan("10d"))
        self.assertTrue(Timespan("10d") < Timestamp(datetime.datetime.now()))

        with self.assertRaises(TypeError):
            _ = Boolean(True) < True
        with self.assertRaises(TypeError):
            _ = Boolean(True) < HandshakeMessage  # Not a data type

        self.assertTrue(Boolean(True) < Count(10))

    def test_unserialize_not_json(self):
        data = b"\x00\x00\x00"

        with self.assertRaisesRegex(TypeError, "cannot parse JSON data"):
            _ = unserialize(data)
        with self.assertRaisesRegex(TypeError, "cannot parse JSON data"):
            _ = Count.unserialize(data)

    def test_unserialize_invalid_json(self):
        data = b"[ 1,2,3 ]"
        with self.assertRaisesRegex(TypeError, "invalid data layout"):
            _ = unserialize(data)
        with self.assertRaisesRegex(TypeError, "invalid data layout"):
            _ = Count.unserialize(data)

        data = b'{ "data": "foobar" }'
        with self.assertRaisesRegex(TypeError, "unrecognized Broker type"):
            _ = unserialize(data)
        with self.assertRaisesRegex(TypeError, "invalid data layout"):
            _ = Count.unserialize(data)

        data = b'{ "data": "foobar", "@data-type": "count" }'
        with self.assertRaisesRegex(TypeError, "invalid data for Count"):
            _ = Count.unserialize(data)

    def test_container_from_broker(self):
        s = Set.from_broker({"data": [{"@data-type": "string", "data": "s"}]})
        self.assertEqual(1, len(s))

        v = Vector.from_broker({"data": [{"@data-type": "string", "data": "s"}]})
        self.assertEqual(1, len(v))

        t = Table.from_broker(
            {
                "data": [
                    {
                        "key": {"@data-type": "string", "data": "s"},
                        "value": {
                            "@data-type": "integer",
                            "data": "42",
                        },
                    },
                ],
            },
        )
        self.assertEqual(1, len(t))



"""This verifies zeek-client invocations."""

import io
import os
import re
import shutil
import subprocess
import tempfile
import unittest
from contextlib import contextmanager

import zeekclient as zc

TESTS = os.path.dirname(os.path.realpath(__file__))
ROOT = os.path.normpath(os.path.join(TESTS, ".."))


# A context guard to switch the current working directory.
# With 3.11 this can go and become contextlib.chdir():
@contextmanager
def setdir(path):
    origin = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


class TestCliInvocation(unittest.TestCase):
    # This invokes the zeek-client toplevel script.

    def test_help(self):
        cproc = subprocess.run(
            ["zeek-client", "--help"],
            check=True,
            capture_output=True,
        )
        self.assertEqual(cproc.returncode, 0)

    def test_show_settings(self):
        cproc = subprocess.run(
            ["zeek-client", "show-settings"],
            check=True,
            capture_output=True,
        )
        self.assertEqual(cproc.returncode, 0)


class TestBundledCliInvocation(unittest.TestCase):
    # Verify that zeek-client finds its package in Zeek-bundled install, where
    # the package will not be in the usual Python search path. As we add more
    # system-level testing, this may become a btest too, but for we stick to
    # Python. Most system-level testing happens in the zeek-testing-cluster
    # external testsuite.
    @unittest.skipUnless(
        shutil.which("cmake") and shutil.which("make"),
        "needs both cmake and make in the system path",
    )
    def test_bundled_install(self):
        with tempfile.TemporaryDirectory() as tmpdir, setdir(tmpdir):
            # Configure the package via cmake with a Python module directory, as
            # Zeek would do. Do this from the temp directory we're now in ...
            cproc = subprocess.run(
                [
                    "cmake",
                    "-D",
                    f'PY_MOD_INSTALL_DIR={os.path.join(tmpdir, "python")}',
                    f"--install-prefix={tmpdir}",
                    ROOT,
                ],
                check=True,
                capture_output=True,
            )

            # ... and install there too, into local bin/ and python/ dirs.
            cproc = subprocess.run(["make", "install"], check=True, capture_output=True)

            # We should now be able to run "./bin/zeek-client --help".
            cproc = subprocess.run(
                [os.path.join(tmpdir, "bin", "zeek-client"), "--help"],
                capture_output=True,
                check=False,
            )
            if cproc.returncode != 0:
                print("==== STDOUT ====")
                print(cproc.stdout.decode("utf-8"))
                print("==== STDERR ====")
                print(cproc.stderr.decode("utf-8"))
                self.fail("zeek-client invocation failed")


class TestCliBasics(unittest.TestCase):
    def test_create_controller(self):
        # We mock create_controller() below, so use this class to test it:
        res = zc.cli.create_controller()
        self.assertIsNotNone(res)
        self.assertIsNotNone(res.controller_broker_id)


class TestCli(unittest.TestCase):
    # This tests the zeekclient.cli module. Most commands in that module create
    # a controller object, so we mock out its generation so we can enqueue the
    # various response events in its websocket.
    def setUp(self):
        # For capturing log writes done by zeekclient code
        self.logbuf = io.StringIO()
        zc.logs.configure(verbosity=2, stream=self.logbuf)

        self.controller = zc.controller.Controller()

        def mock_create_controller():
            self.controller.connect()
            return self.controller

        self.orig_create_controller = zc.cli.create_controller
        zc.cli.create_controller = mock_create_controller

        def mock_make_uuid(_prefix=""):
            return "mocked-reqid-00000"

        self.orig_make_uuid = zc.utils.make_uuid
        zc.controller.make_uuid = mock_make_uuid
        zc.types.make_uuid = mock_make_uuid
        zc.utils.make_uuid = mock_make_uuid

        # Capture regular writes made by the commands,
        # and let us adjust stdin:
        zc.cli.STDOUT = io.StringIO()
        zc.cli.STDIN = io.StringIO()

    def tearDown(self):
        zc.cli.create_controller = self.orig_create_controller

        zc.controller.make_uuid = self.orig_make_uuid
        zc.types.make_uuid = self.orig_make_uuid
        zc.utils.make_uuid = self.orig_make_uuid

    def assertLogLines(self, *patterns):  # noqa: N802
        buflines = self.logbuf.getvalue().split("\n")
        todo = list(patterns)
        for line in buflines:
            if todo and re.search(todo[0], line) is not None:
                todo.pop(0)
        msg = None
        if todo:
            msg = f"log pattern '{todo[0]}' not found; have:\n{self.logbuf.getvalue().strip()}"
        self.assertEqual(len(todo), 0, msg)

    def enqueue_response_event(self, event):
        msg = zc.brokertypes.DataMessage("dummy/topic", event.to_brokertype())
        self.controller.wsock.mock_recv_queue.append(msg.serialize())

    def mock_no_controller(self, inargs):
        # This fakes the scenario where connecting to the controller fails.
        parser = zc.cli.create_parser()
        args = parser.parse_args(inargs)

        def mock_create_controller():
            self.controller.wsock.mock_connect_exc = OSError()
            self.controller.connect()

        zc.cli.create_controller = mock_create_controller

        self.assertEqual(args.run_cmd(args), 1)
        self.assertLogLines("error: socket error in connect()")

    def mock_no_response(self, inargs):
        # This fakes the scenario where an established connection to the
        # controller starts experiencing trouble.
        parser = zc.cli.create_parser()
        args = parser.parse_args(inargs)

        def mock_create_controller():
            self.controller.connect()
            self.controller.wsock.mock_recv_exc = OSError()
            return self.controller

        zc.cli.create_controller = mock_create_controller

        self.assertEqual(args.run_cmd(args), 1)
        self.assertLogLines("error: no response received")

    def test_cmd_deploy_no_controller(self):
        self.mock_no_controller(["deploy"])

    def test_cmd_deploy_no_response(self):
        self.mock_no_response(["deploy"])

    def test_cmd_deploy(self):
        node_outputs = zc.types.NodeOutputs("problems on stdout", "problems on stderr")

        self.enqueue_response_event(
            zc.events.DeployResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.brokertypes.Vector(
                    [
                        zc.types.Result(
                            "reqid-0001",
                            data=zc.brokertypes.String("reqid-config-id"),
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0002",
                            instance="instance1",
                            node="manager",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0003",
                            instance="instance1",
                            node="logger",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0004",
                            success=False,
                            instance="instance1",
                            node="worker1",
                            data=node_outputs.to_brokertype(),
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0005",
                            success=False,
                            instance="instance1",
                            error="uh-oh",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0006",
                            instance="instance1",
                        ).to_brokertype(),
                        zc.types.Result("reqid-0007").to_brokertype(),
                    ],
                ),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["deploy"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_deploy)
        self.assertEqual(args.run_cmd(args), 1)
        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """{
  "errors": [
    "uh-oh"
  ],
  "results": {
    "id": "reqid-config-id",
    "nodes": {
      "logger": {
        "instance": "instance1",
        "success": true
      },
      "manager": {
        "instance": "instance1",
        "success": true
      },
      "worker1": {
        "instance": "instance1",
        "stderr": "problems on stderr",
        "stdout": "problems on stdout",
        "success": false
      }
    }
  }
}
""",
        )

    def test_cmd_get_config_no_controller(self):
        self.mock_no_controller(["get-config"])

    def test_cmd_get_config_no_response(self):
        self.mock_no_response(["get-config"])

    def test_cmd_get_config_as_json(self):
        config = zc.types.Configuration()
        config.instances.append(zc.types.Instance("instance1"))
        config.nodes.append(
            zc.types.Node("worker1", "instance1", zc.types.ClusterRole.WORKER),
        )

        self.enqueue_response_event(
            zc.events.GetConfigurationResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.types.Result(
                    "reqid-0001",
                    data=config.to_brokertype(),
                ).to_brokertype(),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["get-config", "--as-json"])

        # cmd_get_config closes the output handle as part of its regular
        # processing. After a close, the contents of a StringIO object vanish,
        # so we prevent this:
        zc.cli.STDOUT.close = lambda: None

        self.assertEqual(args.run_cmd, zc.cli.cmd_get_config)
        self.assertEqual(args.run_cmd(args), 0)

        zc.cli.STDOUT.flush()

        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """{
  "id": "mocked-reqid-00000",
  "instances": [
    {
      "name": "instance1"
    }
  ],
  "nodes": [
    {
      "cpu_affinity": null,
      "env": {},
      "instance": "instance1",
      "interface": null,
      "metrics_port": null,
      "name": "worker1",
      "options": null,
      "port": null,
      "role": "WORKER",
      "scripts": null
    }
  ]
}
""",
        )

    def test_cmd_get_config_as_ini(self):
        config = zc.types.Configuration()
        config.instances.append(zc.types.Instance("instance1"))
        config.nodes.append(
            zc.types.Node("worker1", "instance1", zc.types.ClusterRole.WORKER),
        )

        self.enqueue_response_event(
            zc.events.GetConfigurationResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.types.Result(
                    "reqid-0001",
                    data=config.to_brokertype(),
                ).to_brokertype(),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["get-config"])

        # cmd_get_config closes the output handle as part of its regular
        # processing. After a close, the contents of a StringIO object vanish,
        # so we prevent this:
        zc.cli.STDOUT.close = lambda: None

        self.assertEqual(args.run_cmd, zc.cli.cmd_get_config)
        self.assertEqual(args.run_cmd(args), 0)

        zc.cli.STDOUT.flush()

        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """[instances]
instance1

[worker1]
instance = instance1
role = WORKER

""",
        )

    def test_cmd_get_config_failure(self):
        self.enqueue_response_event(
            zc.events.GetConfigurationResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.types.Result(
                    "reqid-0001",
                    success=False,
                    error="uh-oh",
                ).to_brokertype(),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["get-config"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_get_config)
        self.assertEqual(args.run_cmd(args), 1)

    def test_cmd_get_config_no_data(self):
        self.enqueue_response_event(
            zc.events.GetConfigurationResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.types.Result("reqid-0001").to_brokertype(),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["get-config"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_get_config)
        self.assertEqual(args.run_cmd(args), 1)

    def test_cmd_get_id_value_no_controller(self):
        self.mock_no_controller(["get-id-value", "Foo:id"])

    def test_cmd_get_id_value_no_response(self):
        self.mock_no_response(["get-id-value", "Foo:id"])

    def test_cmd_get_id_value(self):
        self.enqueue_response_event(
            zc.events.GetIdValueResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.brokertypes.Vector(
                    [
                        zc.types.Result(
                            "reqid-0001",
                            data=zc.brokertypes.String('"a-value"'),
                            node="worker1",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0002",
                            data=zc.brokertypes.String('"b-value"'),
                            node="worker2",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0003",
                            success=False,
                            error="that did not work",
                            node="worker2",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0004",
                            data=zc.brokertypes.Count(10),
                            node="worker2",
                        ).to_brokertype(),
                    ],
                ),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["get-id-value", "Foo:id"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_get_id_value)
        self.assertEqual(args.run_cmd(args), 1)
        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """{
  "errors": [
    {
      "error": "that did not work",
      "source": "worker2"
    },
    {
      "error": "invalid result data type {\\"@data-type\\": \\"count\\", \\"data\\": 10}",
      "source": "worker2"
    }
  ],
  "results": {
    "worker1": "a-value",
    "worker2": "b-value"
  }
}
""",
        )

    def test_cmd_get_instances_no_controller(self):
        self.mock_no_controller(["get-instances"])

    def test_cmd_get_instances_no_response(self):
        self.mock_no_response(["get-instances"])

    def test_cmd_get_instances(self):
        self.enqueue_response_event(
            zc.events.GetInstancesResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.types.Result(
                    "reqid-0001",
                    data=zc.brokertypes.Vector(
                        [
                            zc.types.Instance(
                                "instance1",
                                "10.0.0.1",
                                123,
                            ).to_brokertype(),
                            zc.types.Instance(
                                "instance2",
                                "10.0.0.2",
                                234,
                            ).to_brokertype(),
                        ],
                    ),
                ).to_brokertype(),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["get-instances"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_get_instances)
        self.assertEqual(args.run_cmd(args), 0)
        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """{
  "instance1": {
    "host": "10.0.0.1",
    "port": 123
  },
  "instance2": {
    "host": "10.0.0.2",
    "port": 234
  }
}
""",
        )

    def test_cmd_get_nodes_no_controller(self):
        self.mock_no_controller(["get-nodes"])

    def test_cmd_get_nodes_no_response(self):
        self.mock_no_response(["get-nodes"])

    def test_cmd_get_nodes(self):
        self.enqueue_response_event(
            zc.events.GetNodesResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.brokertypes.Vector(
                    [
                        zc.types.Result(
                            "reqid-0001",
                            instance="instance1",
                            data=zc.brokertypes.Vector(
                                [
                                    zc.types.NodeStatus(
                                        "manager",
                                        zc.types.State.RUNNING,
                                        zc.types.ManagementRole.NONE,
                                        zc.types.ClusterRole.MANAGER,
                                        12345,
                                        2200,
                                    ).to_brokertype(),
                                    zc.types.NodeStatus(
                                        "logger",
                                        zc.types.State.RUNNING,
                                        zc.types.ManagementRole.NONE,
                                        zc.types.ClusterRole.LOGGER,
                                        12346,
                                        2201,
                                    ).to_brokertype(),
                                ],
                            ),
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0002",
                            instance="instance2",
                            data=zc.brokertypes.Vector(
                                [
                                    zc.types.NodeStatus(
                                        "worker1",
                                        zc.types.State.RUNNING,
                                        zc.types.ManagementRole.NONE,
                                        zc.types.ClusterRole.WORKER,
                                        23456,
                                    ).to_brokertype(),
                                    zc.types.NodeStatus(
                                        "worker2",
                                        zc.types.State.RUNNING,
                                        zc.types.ManagementRole.NONE,
                                        zc.types.ClusterRole.WORKER,
                                        23457,
                                    ).to_brokertype(),
                                ],
                            ),
                        ).to_brokertype(),
                        # These cover various error conditions
                        zc.types.Result(
                            "reqid-0003",
                            success=False,
                            instance="instance3",
                            error="uh-oh",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0003",
                            instance="instance4",
                        ).to_brokertype(),
                    ],
                ),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["get-nodes"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_get_nodes)
        self.assertEqual(args.run_cmd(args), 1)
        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """{
  "errors": [
    {
      "error": "uh-oh",
      "source": "instance3"
    },
    {
      "error": "result does not contain node status data",
      "source": "instance4"
    }
  ],
  "results": {
    "instance1": {
      "logger": {
        "cluster_role": "LOGGER",
        "mgmt_role": null,
        "pid": 12346,
        "port": 2201,
        "state": "RUNNING"
      },
      "manager": {
        "cluster_role": "MANAGER",
        "mgmt_role": null,
        "pid": 12345,
        "port": 2200,
        "state": "RUNNING"
      }
    },
    "instance2": {
      "worker1": {
        "cluster_role": "WORKER",
        "mgmt_role": null,
        "pid": 23456,
        "state": "RUNNING"
      },
      "worker2": {
        "cluster_role": "WORKER",
        "mgmt_role": null,
        "pid": 23457,
        "state": "RUNNING"
      }
    }
  }
}
""",
        )

    def test_cmd_restart_no_controller(self):
        self.mock_no_controller(["restart"])

    def test_cmd_restart_no_response(self):
        self.mock_no_response(["restart"])

    def test_cmd_restart(self):
        self.enqueue_response_event(
            zc.events.RestartResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.brokertypes.Vector(
                    [
                        zc.types.Result(
                            "reqid-0001",
                            instance="instance1",
                            node="manager",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0002",
                            instance="instance1",
                            node="logger",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0003",
                            instance="instance2",
                            node="worker1",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0004",
                            instance="instance2",
                            node="worker2",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0004",
                            success=False,
                            node="worker3",
                            error="unknown node",
                        ).to_brokertype(),
                    ],
                ),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["restart"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_restart)
        self.assertEqual(args.run_cmd(args), 1)
        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """{
  "errors": [
    {
      "error": "unknown node",
      "source": "worker3"
    }
  ],
  "results": {
    "logger": true,
    "manager": true,
    "worker1": true,
    "worker2": true
  }
}
""",
        )

    def test_cmd_stage_config_no_controller(self):
        self.mock_no_controller(["stage-config", "-"])

    def test_cmd_stage_config_no_response(self):
        self.mock_no_response(["stage-config", "-"])

    def test_cmd_stage_config(self):
        self.enqueue_response_event(
            zc.events.StageConfigurationResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.brokertypes.Vector(
                    [
                        zc.types.Result(
                            "reqid-0001",
                            data=zc.brokertypes.String("reqid-config-id"),
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0002",
                            success=False,
                            error="uh-oh",
                        ).to_brokertype(),
                    ],
                ),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["stage-config", "-"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_stage_config)
        self.assertEqual(args.run_cmd(args), 1)
        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """{
  "errors": [
    "uh-oh"
  ],
  "results": {
    "id": "reqid-config-id"
  }
}
""",
        )

    def test_cmd_deploy_config(self):
        self.enqueue_response_event(
            zc.events.StageConfigurationResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.brokertypes.Vector(
                    [
                        zc.types.Result(
                            "reqid-0001",
                            data=zc.brokertypes.String("reqid-config-id"),
                        ).to_brokertype(),
                    ],
                ),
            ),
        )

        self.enqueue_response_event(
            zc.events.DeployResponse(
                zc.brokertypes.String(zc.utils.make_uuid()),
                zc.brokertypes.Vector(
                    [
                        zc.types.Result(
                            "reqid-0001",
                            data=zc.brokertypes.String("reqid-config-id"),
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0002",
                            instance="instance1",
                            node="manager",
                        ).to_brokertype(),
                        zc.types.Result(
                            "reqid-0003",
                            instance="instance1",
                            node="logger",
                        ).to_brokertype(),
                    ],
                ),
            ),
        )

        parser = zc.cli.create_parser()
        args = parser.parse_args(["deploy-config", "-"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_deploy_config)
        self.assertEqual(args.run_cmd(args), 0)
        self.assertEqual(
            zc.cli.STDOUT.getvalue(),
            """{
  "errors": [],
  "results": {
    "id": "reqid-config-id",
    "nodes": {
      "logger": {
        "instance": "instance1",
        "success": true
      },
      "manager": {
        "instance": "instance1",
        "success": true
      }
    }
  }
}
""",
        )

    def test_show_settings(self):
        parser = zc.cli.create_parser()
        args = parser.parse_args(["show-settings"])

        self.assertEqual(args.run_cmd, zc.cli.cmd_show_settings)
        self.assertEqual(args.run_cmd(args), 0)

        # The output should be an ini file that we can load back into our
        # configuration:
        zc.CONFIG.read_string(zc.cli.STDOUT.getvalue())



from setuptools import setup


def get_version():
    with open("VERSION", encoding="utf-8") as version:
        return version.read().replace("-", ".dev", 1).strip()


setup(
    version=get_version(),
)



"""This module provides compatibility goop for OpenSSL."""

import os.path
import ssl

from .config import CONFIG


def get_context():  # pragma: no cover
    """Returns an ssl.SSLContext for TLS-protected websocket communication.

    This is a helper for communication protected with SSL, with or without
    authentication. This mirrors Broker's default mode of operation:
    SSL-protected, but without validation/authentication.

    This may raise ssl.SSLError for SSL problems, and OSError if any cert/key
    files cannot be loaded.

    authenticate: if False (default), configures the context to disable
        validation and permit unathenticated ciphersuites.

    Returns: ssl.SSLContext.
    """
    # Newer OpenSSL versions prefer PROTOCOL_TLS_CLIENT over PROTOCOL_TLS, so
    # see if the former is available.
    if hasattr(ssl, "PROTOCOL_TLS_CLIENT"):
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    else:
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLS)

    ssl_certificate = CONFIG.get("ssl", "certificate") or None
    ssl_keyfile = CONFIG.get("ssl", "keyfile") or None
    ssl_cafile = CONFIG.get("ssl", "cafile") or None
    ssl_capath = CONFIG.get("ssl", "capath") or None
    ssl_passphrase = CONFIG.get("ssl", "passphrase") or None

    if ssl_certificate and not os.path.isfile(ssl_certificate):
        raise FileNotFoundError(f'SSL certificate file "{ssl_certificate}" not found')
    if ssl_keyfile and not os.path.isfile(ssl_keyfile):
        raise FileNotFoundError(f'SSL private key file "{ssl_keyfile}" not found')
    if ssl_cafile and not os.path.isfile(ssl_cafile):
        raise FileNotFoundError(f'SSL trusted CAs file "{ssl_cafile}" not found')
    if ssl_capath and not os.path.isdir(ssl_capath):
        raise FileNotFoundError(f'SSL trusted CAs path "{ssl_capath}" not found')

    if not ssl_certificate:
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
        # This mirrors the selection in Broker, and depends on the OpenSSL version:
        try:
            ctx.set_ciphers("AECDH-AES256-SHA@SECLEVEL=0")
        except ssl.SSLError:
            ctx.set_ciphers("AECDH-AES256-SHA")
    else:
        # This too mirrors Broker:
        ctx.set_ciphers("HIGH:!aNULL:!MD5")
        ctx.load_cert_chain(ssl_certificate, ssl_keyfile, ssl_passphrase)
        if ssl_cafile or ssl_capath:
            ctx.load_verify_locations(cafile=ssl_cafile, capath=ssl_capath)

    return ctx


def get_websocket_sslopt():
    """Returns a TLS options dict for websocket-client.

    The resulting dict is suitable for the websocket.WebSocket()
    constructor. It's required for older websocket-client versions that don't
    yet support passing an SSL context explicitly. This can go when everyone can
    easily use websocket-client >= 1.2.2.
    """
    ssl_certificate = CONFIG.get("ssl", "certificate") or None
    ssl_keyfile = CONFIG.get("ssl", "keyfile") or None
    ssl_cafile = CONFIG.get("ssl", "cafile") or None
    ssl_capath = CONFIG.get("ssl", "capath") or None
    ssl_passphrase = CONFIG.get("ssl", "passphrase") or None

    if ssl_certificate and not os.path.isfile(ssl_certificate):
        raise FileNotFoundError(f'SSL certificate file "{ssl_certificate}" not found')
    if ssl_keyfile and not os.path.isfile(ssl_keyfile):
        raise FileNotFoundError(f'SSL private key file "{ssl_keyfile}" not found')
    if ssl_cafile and not os.path.isfile(ssl_cafile):
        raise FileNotFoundError(f'SSL trusted CAs file "{ssl_cafile}" not found')
    if ssl_capath and not os.path.isdir(ssl_capath):
        raise FileNotFoundError(f'SSL trusted CAs path "{ssl_capath}" not found')

    # SSL options as understood by websocket-client
    sslopt = {}

    if not ssl_certificate:
        sslopt["cert_reqs"] = ssl.CERT_NONE
        if ssl.OPENSSL_VERSION_NUMBER >= 0x10100000:
            sslopt["ciphers"] = "AECDH-AES256-SHA@SECLEVEL=0"
        else:
            sslopt["ciphers"] = "AECDH-AES256-SHA"
    else:
        sslopt["ciphers"] = "HIGH:!aNULL:!MD5"

        if ssl_certificate:
            sslopt["certfile"] = ssl_certificate
        if ssl_keyfile:
            sslopt["keyfile"] = ssl_keyfile
        if ssl_passphrase:
            sslopt["password"] = ssl_passphrase
        if ssl_cafile:
            sslopt["ca_certs"] = ssl_cafile
        if ssl_capath:
            sslopt["ca_cert_path"] = ssl_capath

    return sslopt



"""This verifies zeekclient.controller.Controller's behavior."""

import io
import os
import re
import ssl
import unittest

import websocket

import zeekclient

TESTS = os.path.dirname(os.path.realpath(__file__))


class TestController(unittest.TestCase):
    def setUp(self):
        # A buffer receiving any created log messages, for validation. We could
        # also assertLogs(), but with the latter it's more work to get exactly
        # the output the user would see.
        self.logbuf = io.StringIO()
        zeekclient.logs.configure(verbosity=2, stream=self.logbuf)
        zeekclient.CONFIG.reset()

    def assertLogLines(self, *patterns):  # noqa: N802
        buflines = self.logbuf.getvalue().split("\n")
        todo = list(patterns)
        for line in buflines:
            if todo and re.search(todo[0], line) is not None:
                todo.pop(0)
        msg = None
        if todo:
            msg = f"log pattern '{todo[0]}' not found; have:\n{self.logbuf.getvalue().strip()}"
        self.assertEqual(len(todo), 0, msg)

    def test_connect_successful(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        self.assertEqual(
            controller.controller_broker_id,
            controller.wsock.mock_broker_id,
        )
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            "info: peered with controller 127.0.0.1:2149",
        )

    def test_connect_successful_custom_controller(self):
        controller = zeekclient.controller.Controller("example.com", 1234)
        self.assertTrue(controller.connect())
        self.assertLogLines(
            "info: connecting to controller example.com:1234",
            "info: peered with controller example.com:1234",
        )

    def test_connect_successful_no_tls(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        self.assertTrue(controller.wsock.mock_url.startswith("ws://"))
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            "info: peered with controller 127.0.0.1:2149",
        )

    def test_connect_ssldisable_deprecation(self):
        # The old default setting, to enable unvalidated TLS:
        zeekclient.CONFIG.set("ssl", "disable", "false")
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        self.assertTrue(controller.wsock.mock_url.startswith("ws://"))
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            "info: peered with controller 127.0.0.1:2149",
        )

    def test_connect_successful_authenticated_tls(self):
        zeekclient.CONFIG.set("ssl", "enable", "true")
        zeekclient.CONFIG.set(
            "ssl",
            "certificate",
            os.path.join(TESTS, "certs", "cert.1.pem"),
        )
        zeekclient.CONFIG.set(
            "ssl",
            "keyfile",
            os.path.join(TESTS, "certs", "key.1.pem"),
        )
        zeekclient.CONFIG.set("ssl", "cafile", os.path.join(TESTS, "certs", "ca.pem"))
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        self.assertTrue(controller.wsock.mock_url.startswith("wss://"))
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            "info: peered with controller 127.0.0.1:2149",
        )

    def test_connect_successful_authenticated_tls_pw(self):
        zeekclient.CONFIG.set("ssl", "enable", "true")
        zeekclient.CONFIG.set(
            "ssl",
            "certificate",
            os.path.join(TESTS, "certs", "cert.1.pem"),
        )
        zeekclient.CONFIG.set(
            "ssl",
            "keyfile",
            os.path.join(TESTS, "certs", "key.1.pem"),
        )
        zeekclient.CONFIG.set("ssl", "cafile", os.path.join(TESTS, "certs", "ca.pem"))
        zeekclient.CONFIG.set("ssl", "passphrase", "12345")
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            "info: peered with controller 127.0.0.1:2149",
        )

    def test_connect_successful_authenticated_tls_file_config_errors(self):
        for error in ("certificate", "keyfile", "cafile"):
            zeekclient.CONFIG.set("ssl", "enable", "true")
            zeekclient.CONFIG.set(
                "ssl",
                "certificate",
                os.path.join(TESTS, "certs", "cert.1.pem"),
            )
            zeekclient.CONFIG.set(
                "ssl",
                "keyfile",
                os.path.join(TESTS, "certs", "key.1.pem"),
            )
            zeekclient.CONFIG.set(
                "ssl",
                "cafile",
                os.path.join(TESTS, "certs", "ca.pem"),
            )
            # Break the configuration:
            zeekclient.CONFIG.set("ssl", error, "not-a-file")

            with self.assertRaises(zeekclient.controller.ConfigError):
                _ = zeekclient.controller.Controller()

    def test_connect_fails_with_refused(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_connect_exc = ConnectionRefusedError()
        # Dial down attempts and waits to make this fast:
        zeekclient.CONFIG.set("client", "peering_attempts", "2")
        zeekclient.CONFIG.set("client", "peering_retry_delay_secs", "0.1")
        self.assertFalse(controller.connect())
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            r"error: websocket connection to 127.0.0.1:2149 timed out in connect\(\)",
        )

    def test_connect_fails_with_timeout(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_connect_exc = websocket.WebSocketTimeoutException(
            "connection timed out",
        )
        # Dial down attempts and waits to make this fast:
        zeekclient.CONFIG.set("client", "peering_attempts", "2")
        zeekclient.CONFIG.set("client", "peering_retry_delay_secs", "0.1")
        self.assertFalse(controller.connect())
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            r"error: websocket connection to 127.0.0.1:2149 timed out in connect\(\)",
        )

    def test_connect_fails_with_websocket_error(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_connect_exc = websocket.WebSocketException("uh-oh")
        self.assertFalse(controller.connect())
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            r"error: websocket error in connect\(\) with controller 127.0.0.1:2149: uh-oh",
        )

    def test_connect_fails_with_sslerror(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_connect_exc = ssl.SSLError(
            "dummy library version",
            "uh-oh",
        )
        self.assertFalse(controller.connect())
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            r"error: socket TLS error in connect\(\) with controller 127.0.0.1:2149: uh-oh",
        )

    def test_connect_fails_with_oserror(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_connect_exc = OSError("uh-oh")
        self.assertFalse(controller.connect())
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            r"error: socket error in connect\(\) with controller 127.0.0.1:2149: uh-oh",
        )

    def test_connect_fails_with_unknown_error(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_connect_exc = websocket.UnknownError("surprise")
        self.assertFalse(controller.connect())
        self.assertLogLines(
            "info: connecting to controller 127.0.0.1:2149",
            r"error: unexpected error in connect\(\) with controller 127.0.0.1:2149: surprise",
        )

    def test_handshake_fails_with_timeout(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_recv_exc = websocket.WebSocketTimeoutException(
            "connection timed out",
        )
        self.assertFalse(controller.connect())
        self.assertLogLines("error: websocket connection to .+ timed out in handshake")

    def test_handshake_fails_with_oserror(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_recv_exc = OSError("uh-oh")
        self.assertFalse(controller.connect())
        self.assertLogLines(
            "error: socket error in handshake with controller .+: uh-oh",
        )

    def test_handshake_fails_with_unknown_error(self):
        controller = zeekclient.controller.Controller()
        controller.wsock.mock_recv_exc = websocket.UnknownError("surprise")
        self.assertFalse(controller.connect())
        self.assertLogLines(
            "error: unexpected error in handshake with controller .+: surprise",
        )

    def test_handshake_fails_with_protocol_data_error(self):
        controller = zeekclient.controller.Controller()
        # Not a Handshake ACK message:
        controller.wsock.mock_recv_queue = [zeekclient.brokertypes.Count(1).serialize()]
        self.assertFalse(controller.connect())
        self.assertLogLines("error: protocol data error")

    def test_publish(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())

        reqid = zeekclient.utils.make_uuid()
        event = zeekclient.events.GetConfigurationRequest(reqid, True)

        controller.publish(event)

        # The event gets transmitted via the controller object's websocket, so
        # verify it's as expected: a DataMessage containing our event. This is
        # the first message after the initial handshake, so the second in the
        # queue.
        message = zeekclient.brokertypes.DataMessage.unserialize(
            controller.wsock.mock_send_queue[1],
        )
        self.assertEqual(event.to_brokertype().serialize(), message.data.serialize())

    def test_receive(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())

        event = zeekclient.events.GetConfigurationResponse(
            zeekclient.utils.make_uuid(),
            (),
        )

        # Mock an event in the receive queue, so we can receive something:
        controller.wsock.mock_recv_queue.append(
            zeekclient.brokertypes.DataMessage(
                "dummy/topic",
                event.to_brokertype(),
            ).serialize(),
        )

        event, error = controller.receive()

        self.assertIsInstance(event, zeekclient.events.GetConfigurationResponse)
        self.assertEqual(error, "")

    def test_receive_fails_with_protocol_data_error(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        # Not a DataMessage:
        controller.wsock.mock_recv_queue.append('{"type": "data-message"}')
        res, msg = controller.receive()
        self.assertIsNone(res)
        self.assertRegex(
            msg,
            "protocol data error .+: invalid data layout for DataMessage",
        )

    def test_receive_fails_with_timeout(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        controller.wsock.mock_recv_exc = websocket.WebSocketTimeoutException(
            "connection timed out",
        )
        res, msg = controller.receive()
        self.assertIsNone(res)
        self.assertRegex(msg, "websocket connection .+ timed out")

    def test_receive_fails_with_unknown_error(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        controller.wsock.mock_recv_exc = websocket.UnknownError("surprise")
        res, msg = controller.receive()
        self.assertIsNone(res)
        self.assertRegex(msg, "unexpected error .+: surprise")

    def test_receive_fails_with_event_error(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())
        # A DataMessage, but not with an event:
        controller.wsock.mock_recv_queue.append(
            zeekclient.brokertypes.DataMessage(
                "dummy/topic",
                zeekclient.brokertypes.Vector(),
            ).serialize(),
        )
        res, msg = controller.receive()
        self.assertIsNone(res)
        self.assertRegex(msg, "protocol data error .+: invalid event data")

    def test_transact(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())

        reqid = zeekclient.utils.make_uuid()
        event = zeekclient.events.DeployResponse(reqid, ())

        # Mock an event in the receive queue, so we can receive something:
        controller.wsock.mock_recv_queue.append(
            zeekclient.brokertypes.DataMessage(
                "dummy/topic",
                event.to_brokertype(),
            ).serialize(),
        )

        event, error = controller.transact(
            zeekclient.events.DeployRequest,
            zeekclient.events.DeployResponse,
            reqid=reqid,
        )

        self.assertIsInstance(event, zeekclient.events.DeployResponse)
        self.assertEqual(error, "")
        self.assertEqual(event.reqid.to_py(), reqid)

    def test_transact_data_mismatches(self):
        controller = zeekclient.controller.Controller()
        self.assertTrue(controller.connect())

        reqid = zeekclient.utils.make_uuid()

        # Fill the receive queue with events. The first response mismatches in
        # its name, the second in its first argument (not reqid), the third
        # passes.
        events = [
            zeekclient.events.GetConfigurationResponse(reqid, ()),
            zeekclient.events.DeployResponse("xxxx", ()),
            zeekclient.events.DeployResponse(reqid, ()),
        ]

        for evt in events:
            controller.wsock.mock_recv_queue.append(
                zeekclient.brokertypes.DataMessage(
                    "dummy/topic",
                    evt.to_brokertype(),
                ).serialize(),
            )

        event, error = controller.transact(
            zeekclient.events.DeployRequest,
            zeekclient.events.DeployResponse,
            reqid=reqid,
        )

        self.assertIsInstance(event, zeekclient.events.DeployResponse)
        self.assertEqual(error, "")
        self.assertEqual(event.reqid.to_py(), reqid)
